__all__ = [
    "Pluralize",
    "humanize",
    "api_route",
]

import functools

import pendulum
import tornado.web


class Pluralize(tornado.web.UIModule):
    """Pluralize a string based on a value.

    You Must set `Pluralize` as a valid UIModule
    In `ui_modules` setting like so.

    ui_modules=dict(
        'pluralize': Pluralize,
    )
    """

    IRREGULAR_NOUNS = {
        "woman": "women",
        "man": "men",
        "child": "children",
        "tooth": "teeth",
        "foot": "feet",
        "person": "people",
        "leaf": "leaves",
        "mouse": "mice",
        "goose": "geese",
        "half": "halves",
        "knife": "knives",
        "wife": "wives",
        "life": "lives",
        "elf": "elves",
        "loaf": "loaves",
        "potato": "potatoes",
        "tomato": "tomatoes",
        "cactus": "cacti",
        "focus": "foci",
        "fungus": "fungi",
        "nucleus": "nuclei",
        "syllabus": "syllabuses",
        "analysis": "analyses",
        "diagnosis": "diagnoses",
        "oasis": "oases",
        "thesis": "theses",
        "crisis": "crises",
        "phenomenon": "phenomena",
        "criterion": "criteria",
        "datum": "data",
    }

    SAME_FORMS = {
        "sheep": "sheep",
        "fish": "fish",
        "deer": "deer",
        "species": "species",
        "aircraft": "aircraft",
    }

    def render(self, word: str, count: int) -> str:
        if count > 1:
            if word in self.IRREGULAR_NOUNS:
                return self.IRREGULAR_NOUNS.get(word)
            elif word in self.SAME_FORMS:
                return self.SAME_FORMS.get(word)
            elif word.endswith(("s", "x", "z", "ch", "sh")):
                return f"{word}es"
            elif word[-2] not in ["o", "u", "e", "i", "a"] and word.endswith(
                "y"
            ):  # noqa E501
                return f"{word[:len(word)-1]}ies"
            else:
                return f"{word}s"
        return word


def humanize(func):
    """Humanize datetime in templates.

    Take a look at example directory.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        field = func(*args, **kwargs)
        # replace naive datetime tzinfo with pendulum tzinfo
        # naive datetime(is not comparable)
        # timezone aware datetime(is comparable)
        local_timezone = pendulum.tz.get_local_timezone()
        field = field.replace(tzinfo=local_timezone)
        result = pendulum.now().diff_for_humans(field)
        result = result.split(" ")
        for word in result:
            if word == "after":
                result.remove(word)
        return " ".join([w for w in result])

    return wrapper


class Route(object):
    """Usernado API router class."""

    urls = []

    def __call__(self, url: str, name: str = None):
        def wrapper(cls):
            self.urls.append(
                tornado.web.URLSpec(
                    url, cls, name=name if name else cls.__name__.lower()
                )
            )
            return cls

        return wrapper


api_route = Route()



__all__ = [
    "UserAlreadyExistError",
    "UserDoesNotExistError",
    "UnsupportedUserModelError",
    "PermissionDeniedError",
    "DataMalformedOrNotProvidedError",
]

from torntriplets.web import (
    UserAlreadyExistError,
    UserDoesNotExistError,
    UnsupportedUserModelError,
    PermissionDeniedError,
)
from torntriplets.api import DataMalformedOrNotProvidedError



from typing import Any, Dict, Optional

import tornado.escape
import tornado.web

from usernado.torntriplets.base import BaseHandler


class BaseValidationError(ValueError):
    pass


class DataMalformedOrNotProvidedError(BaseValidationError):
    pass


class APIHandler(BaseHandler):
    def get_json_argument(
        self,
        name: str,
        default: Optional[str] = None,
    ) -> str:
        """Get json argument from current request.

        :param name: Name of the argument
        :type name: str
        :param default: Default value for argument if not presented,
         defaults to None
        :type default: str, optional
        :raises DataMalformedOrNotProvidedError:
        :return: Particular JSON argument that comes with current request
        :rtype: str
        """
        try:
            raw_data = self.request.body.decode().replace("'", '"')
        except Exception:
            raise DataMalformedOrNotProvidedError
        else:
            json_data = tornado.escape.json_decode(raw_data)
            return json_data.get(name, default)

    def get_json_arguments(self) -> Dict[Any, Any]:
        """Get all json arguments from current request.

        :raises DataMalformedOrNotProvidedError:
        :return: All JSON argument that comes with current request
        :rtype: Dict[Any, Any]
        """
        try:
            raw_data = self.request.body.decode().replace("'", '"')
        except Exception:
            raise DataMalformedOrNotProvidedError
        else:
            json_data = tornado.escape.json_decode(raw_data)
            return json_data



__version__ = "0.2.6"
__all__ = ["Usernado"]

from usernado.torntriplets import APIHandler, WebHandler, WebSocketHandler


class Usernado:
    API = APIHandler
    Web = WebHandler
    WebSocket = WebSocketHandler



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['usernado', 'usernado.torntriplets']

package_data = \
{'': ['*']}

install_requires = \
['pendulum>=2.1.2,<3.0.0', 'tornado-debugger>=0.9.0,<0.10.0']

setup_kwargs = {
    'name': 'usernado',
    'version': '0.2.6',
    'description': 'Usernado is a Tornado Extension to Make Life Easier.',
    'long_description': '<a id="top"></a>\n<br />\n\n<div align="center">\n  <h1>Usernado</h1>\n  <p align="center">\n    Makes it Easy to Manage Tornado :tornado: Applications\n    <br />\n    <a href="#"><strong>Explore the docs »</strong></a>\n    <br />\n    <br />\n    <a href="https://github.com/reganto/Usernado/issues"><img src="https://img.shields.io/github/issues/reganto/usernado"></a> <a href="https://github.com/reganto/usernado/blob/master/LICENSE.txt"><img src="https://img.shields.io/github/license/reganto/usernado"></a>  <a href="https://badge.fury.io/py/usernado"><img src="https://badge.fury.io/py/usernado.svg" alt="PyPI version" height="18"></a> <a href="https://pepy.tech/project/usernado"><img src="https://pepy.tech/badge/usernado"/></a>\n  </p>\n</div>\n\n<!-- Why Userndo  -->\n\n## Why Usernado\n\nI\'m using 🌪️ every day. I really like it 😍 . Besides of all advantages of Tornado, it\'s not a full-stack framework, and I had to put all the pieces of the puzzle together every day 😩! So this is my attempt to follow DRY(Don\'t Repeat Yourself) principle. This is how the Usernado was born.\n\n<!-- Features -->\n\n## Features\n\n- REST support :zap:\n\n- Websocket easier than ever :zap:\n\n- ORM agnostic authentication :zap:\n\n- Humanize datetime in templates :zap:\n\n- Better exception printer thanks to [tornado-debugger](https://github.com/bhch/tornado-debugger) :zap:\n\n<!-- Getting Started -->\n\n## Installation\n\nInstall either with pip or poetry.\n\n```bash\npip install usernado\n```\n```bash\npoetry add usernado\n```\n\nOr optionally you can install from github using \n```bash \npip install git+https://github.com/reganto/usernado\n```\n\n<!-- USAGE EXAMPLES -->\n\n## Usage\n\n### Example\n\n```python\nfrom usernado import Usernado\n\n\nclass HelloHandler(Usernado.Web):\n    def get(self):\n        self.write("Hello, World!")\n```\n\nFor more examples please Check out [examples](https://github.com/reganto/Usernado/tree/master/example).\n\n<!-- ROADMAP -->\n\n## Roadmap\n\n- [x] Send and broadcast for websockets\n- [x] Abstracted authentication methods\n- [x] Authenticaion methods should return True/False\n- [x] Add diff_for_human (humanize) decorator\n- [x] Add api_route for API handlers\n- [x] Add username & password to test login \n- [x] Add pluralize (str_plural) uimodule\n- [ ] Add third party authentication abstract methods\n- [ ] Add pagination\n\n<!-- CONTACT -->\n\n## Contact\n\nEmail: tell.reganto[at]gmail[dot]com\n\n<p align="right"><a href="#top"><img src="https://raw.githubusercontent.com/DjangoEx/python-engineer-roadmap/main/statics/top.png" width=50 height=50 /></a></p>\n',
    'author': 'Reganto',
    'author_email': 'tell.reganto@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/reganto/usernado',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""If you want to add support for other ORM's:
01- Create a class which implement `IAuth` interface
 with name convention like so: ORMNAMEAuth
02- Override `register` and `login` methods
03- Add model check logic in WebHandler's register
 and login methods.
"""

import hashlib
import secrets
from abc import ABCMeta, abstractmethod
from typing import Optional, Union

from tornado.escape import xhtml_escape

from usernado.torntriplets.base import BaseHandler


class BaseValidationError(ValueError):
    pass


class UserDoesNotExistError(BaseValidationError):
    pass


class UserAlreadyExistError(BaseValidationError):
    pass


class PermissionDeniedError(BaseValidationError):
    pass


class UnsupportedUserModelError(BaseValidationError):
    pass


_SALT = secrets.token_hex()


def _hash_password(password: str, salt: str = _SALT) -> str:
    """Generate hashed password.

    :param password: Incomming password
    :type password: str
    :param salt: A salt to make hashing more strength, defaults to _SALT
    :type salt: str, optional
    :return: A hashed password
    :rtype: str
    """

    password = password.encode("utf-8")
    hashed_password = hashlib.sha512(password + salt.encode()).hexdigest()
    return hashed_password


def _sqlalchemy_session_maker():
    """Generate a session for SQLAlchemy.

    There seems to be a better implementation to do this!"""

    from sqlalchemy import create_engine
    from sqlalchemy.orm import sessionmaker

    import database

    engine = create_engine(database.models.DB)
    Session = sessionmaker(bind=engine)
    session = Session()
    return session


class IAuth(metaclass=ABCMeta):
    """Every ORM specific authentication class MUST implement
    this Interface and override `register` and `login` mehtods.
    """

    @staticmethod
    @abstractmethod
    def register(request, model, username, password):
        NotImplementedError

    @staticmethod
    @abstractmethod
    def login(request, model, username, password):
        NotImplementedError


class PeeweeAuth(IAuth):
    """Peewee specific authentication class."""

    def register(request, model, username, password):
        hashed_password = _hash_password(password)
        user_already_exist = model.select().where(model.username == username).first()

        if user_already_exist:
            raise UserAlreadyExistError(user_already_exist)

        try:
            model.create(username=username, password=hashed_password, salt=_SALT)
        except Exception as e:
            print("Error in user registration proccess: ", e)
        else:
            return True

    def login(request, model, username, password):
        user = None
        try:
            user = model.filter(model.username == username).first()
        except Exception:
            raise UserDoesNotExistError
        else:
            hashed_password = _hash_password(password, salt=user.salt)

        if user and user.password == hashed_password:
            request.set_secure_cookie(
                "username",
                username,
                secure=True,
                httpOnly=True,
            )
            return True
        else:
            raise PermissionDeniedError("Your username or password is NOT correct.")


class SQLAlchemy(IAuth):
    """SQLAlchemy specific authentication class."""

    def register(request, model, username, password):
        hashed_password = _hash_password(password)
        session = _sqlalchemy_session_maker()
        user_already_exist = session.query(model).filter_by(username=username).first()

        if user_already_exist:
            raise UserAlreadyExistError(user_already_exist)

        try:
            user = model(username=username, password=hashed_password, salt=_SALT)
            session.add(user)
            session.commit()
        except Exception as e:
            print("Error in user registration proccess: ", e)
            session.rollback()
        else:
            return True
        finally:
            session.close()

    def login(request, model, username, password):
        session = _sqlalchemy_session_maker()
        user_exist = session.query(model).filter_by(username=username).first()

        if not user_exist:
            raise UserDoesNotExistError("User does not exist.")

        hashed_password = _hash_password(password, salt=user_exist.salt)

        if user_exist and user_exist.password == hashed_password:
            request.set_secure_cookie(
                "username",
                username,
                secure=True,
                httpOnly=True,
            )
            return True
        else:
            raise PermissionDeniedError("Your username or password is NOT correct.")


class WebHandler(BaseHandler):
    def register(self, user_model, username: str, password: str) -> bool:
        """Signup user with provided username and password.

        :param user_model: User model
        :type user_model: SQLAlchemy | Peewee |
        :param username: Username
        :type username: str
        :param password: Password
        :type password: str
        :raises UnsupportedUserModelError: If user_model was not
          an instance of SQLAlchemy or Peewee.
        :return: Return True, If user registration is successful.
        :rtype: bool
        """

        try:
            import peewee

            if issubclass(user_model, peewee.Model):
                return PeeweeAuth.register(
                    request=self,
                    model=user_model,
                    username=username,
                    password=password,
                )
            else:
                raise UnsupportedUserModelError(user_model)
        except (ModuleNotFoundError, UnsupportedUserModelError):
            try:
                import sqlalchemy

                if user_model.metadata:
                    # I've no experience in sqlalchemy so I look for a better
                    # implementation to check if user_model is type of sqlalchemy or not.
                    return SQLAlchemy.register(
                        request=self,
                        model=user_model,
                        username=username,
                        password=password,
                    )
                else:
                    raise UnsupportedUserModelError(user_model)
            except ModuleNotFoundError:
                self.write("<h3>You have to install SQLAlchemy or Peewee first.</h3>")

    def login(self, user_model, username: str, password: str) -> bool:
        """Signin user with provided username and password.

        :param user_model: User model
        :type user_model: SQLAlchemy | Peewee |
        :param username: Username
        :type username: str
        :param password: Password
        :type password: str
        :raises UnsupportedUserModelError: If user_model was not
          an instance of SQLAlchemy or Peewee.
        :return: Return True, if user login is successful.
        :rtype: True
        """

        try:
            import peewee

            if issubclass(user_model, peewee.Model):
                return PeeweeAuth.login(
                    request=self,
                    model=user_model,
                    username=username,
                    password=password,
                )
            else:
                raise UnsupportedUserModelError(user_model)
        except (ModuleNotFoundError, UnsupportedUserModelError):
            try:
                import sqlalchemy

                if user_model.metadata:
                    # I've no experience in sqlalchemy so I look for a better
                    # implementation to check if user_model is type of sqlalchemy or not.
                    return SQLAlchemy.login(
                        request=self,
                        model=user_model,
                        username=username,
                        password=password,
                    )
                else:
                    raise UnsupportedUserModelError(user_model)
            except ModuleNotFoundError:
                self.write("<h3>You have to install SQLAlchemy or Peewee first.</h3>")

    def logout(self) -> None:
        """Logout user."""

        self.clear_cookie("username")

    @property
    def authenticate(self) -> bool:
        """Check if current user is authenticated.

        :rtype: bool
        """

        return bool(self.current_user)

    def get_current_user(self) -> str:
        """To implement user authentication we need to override this method.

        for more information, take a look at Tornado documentation.
        https://www.tornadoweb.org/en/stable/guide/security.html?highlight=get_current_user#user-authentication

        :return: A secure cookie
        :rtype: str
        """

        return self.get_secure_cookie("username")

    def redirect_to_route(self, name: str):
        """Redirect to particular route.

        :param name: Named route
        :type name: str
        """

        self.redirect(self.reverse_url(name))

    def get_escaped_argument(
        self,
        name: str,
        default: Union[None, str],
        strip: bool = True,
    ) -> Optional[str]:
        """Returns the escaped value of the argument with the given name.

        :param name: Name of desired argument
        :type name: str
        :return: Escaped argument
        :rtype: str
        """
        argument = self.get_argument(name, default, strip)
        return xhtml_escape(argument)
